import { readdir } from "node:fs/promises";
import { resolve } from "node:path";

import { discord } from "@config/environment";
import { logger } from "@helpers/logger";
import {
	type AnyInteractionGateway,
	ApplicationCommandTypes,
	Client,
	CommandInteraction,
	Message,
} from "oceanic.js";

const client: Client & { commands: Map<string, Command> } = Object.assign(
	new Client({
		auth: `Bot ${discord.token}`,
		allowedMentions: {
			everyone: false,
			repliedUser: false,
			roles: true,
			users: true,
		},
		defaultImageFormat: "png",
		defaultImageSize: 4096,
		disableCache: false,
		gateway: {
			intents: ["ALL"],
		},
	}),
	{ commands: new Map<string, Command>() },
);

const loadCommands: () => Promise<void> = async () => {
	const commandsPath: string = resolve("src", "commands");
	const commandFiles: string[] = await readdir(commandsPath);

	for (const file of commandFiles) {
		if (!file.endsWith(".ts")) continue;
		const commandModule: Import = await import(resolve(commandsPath, file));

		if (commandModule.default && commandModule.default.name) {
			client.commands.set(commandModule.default.name, {
				...commandModule.default,
				interaction: commandModule.interaction,
				legacy: commandModule.legacy,
			});

			logger.info(`Loaded command: ${commandModule.default.name}`);
		} else {
			logger.warn(`Command file ${file} is missing a valid export.`);
		}
	}

	const globalCommands: Array<{
		name: string;
		description: string;
		options: [];
		type: ApplicationCommandTypes;
	}> = Array.from(client.commands.values()).map((cmd: Command) => ({
		name: cmd.name,
		description: cmd.description,
		options: cmd.options || [],
		type: ApplicationCommandTypes.CHAT_INPUT,
	}));

	await client.application.bulkEditGlobalCommands(globalCommands);
};

client.on("ready", async (): Promise<void> => {
	logger.info(`Ready as ${client.user.tag}`, true);
	logger.info("Loading client.commands...");
	await loadCommands();
});

client.on(
	"interactionCreate",
	async (interaction: AnyInteractionGateway): Promise<void> => {
		if (interaction instanceof CommandInteraction) {
			const command: Command | undefined = client.commands.get(
				interaction.data.name,
			);
			if (command && command.interaction) {
				try {
					await command.interaction(interaction);
				} catch (error) {
					logger.error(
						`Error executing interaction command ${interaction.data.name}:`,
					);
					logger.error(error as Error);
					await interaction.createMessage({
						content: "There was an error executing that command.",
					});
				}
			} else {
				logger.warn(
					`No interaction handler found for ${interaction.data.name}`,
				);
			}
		}
	},
);

client.on("messageCreate", async (message: Message) => {
	if (message.author.bot || !message.content.startsWith(discord.prefix))
		return;

	const args: string[] = message.content
		.slice(discord.prefix.length)
		.trim()
		.split(/\s+/);
	const commandName: string | undefined = args.shift()?.toLowerCase();
	if (!commandName) return;

	const command: Command | undefined = client.commands.get(commandName);
	if (command && command.legacy) {
		try {
			await command.legacy(message);
		} catch (error) {
			logger.error(`Error executing legacy command ${commandName}:`);
			logger.error(error as Error);

			if (message.channel)
				await message.channel.createMessage({
					content: "There was an error executing this command.",
				});
		}
	} else {
		logger.warn(`No legacy handler found for ${commandName}`);
	}
});

client.on("error", (err: string | Error) => {
	logger.error("Client error:");
	logger.error(err);
});

client.connect();
